"""
Module which contains the base classes for the framework
"""

import logging
from typing import Dict, Generator

import requests
from tenacity import retry, stop_after_attempt, wait_fixed, wait_random

from sports.utils import convert_http_to_https, get_id_from_url, join_url


class DataModel:
    """
    Base plain data model mapped to the data model in the server.
    """

    def __init__(self, model_resource):
        self.model_resource = model_resource

    def delete(self) -> None:
        """
        Shortcut to delete a model. Actually, the work is delegated to
        the underlying 'model_resource'
        """
        self.model_resource.delete_remote_resource(self)

    def refresh_from_server(self) -> None:
        """
        Reload field values from the server.
        """
        latest = self.model_resource.get_remote_resource(self)
        for field in self.model_resource.get_fields:
            setattr(self, field, getattr(latest, field))


class RESTServer:
    """
    The base class which is responsible to interactive with REST APIs
    exposed by the server.
    """

    def __init__(
        self, server, api_version="/api/v1/", item_key="results", ssl_verify=True,
    ):
        self._server = server
        self._api_version = api_version
        self._url_root = join_url(self._server, self._api_version)
        self._item_key = item_key
        self._session = requests.Session()
        self._session.verify = ssl_verify

    @property
    def item_key(self) -> str:
        return self._item_key

    @property
    def url_root(self) -> str:
        return self._url_root

    def set_csrf(self, csrf_token: str) -> None:
        """
        Set the CSRF token to include in the headers for requests.
        """
        self._session.headers.update({"X-CSRF-TOKEN": csrf_token})

    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2) + wait_random(0, 2))
    def _raw_get(self, url: str, params=None, headers=None) -> Dict:
        """
        Return an un-interpreted response for a GET of a full URL.
        """
        logging.debug(f"GET REQUEST: {url}")
        response = self._session.get(url, params=params, headers=headers)
        response.raise_for_status()
        logging.debug(f"RESPONSE: {response.status_code}")
        return response

    def get(self, url: str, params=None, headers=None) -> Generator:
        """
        A wrapper method for the internal method '_raw_get', which returns all
        the results.
        It's best to use this method when the queryset is small. For a large
        queryset, use the method 'get_by_page' instead.
        """
        return self._raw_get(url, params=params, headers=headers).json()

    def get_by_page(self, url: str, params=None, headers=None) -> Generator:
        """
        Perform a stream of GET requests by pagination
        """
        if params is None:
            params = {}
        not_done = True
        while not_done:
            resp_data = self._raw_get(url, params=params, headers=headers).json()
            items = resp_data.get(self.item_key)
            url = resp_data["next"]
            not_done = bool(items) and bool(url)
            for value in items:
                yield value

    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2) + wait_random(0, 2))
    def post(self, url, data=None, **kwargs) -> Dict:
        """
        Perform a POST request against the REST server.
        """
        logging.debug(f"POST REQUEST: {url}")
        response = self._session.post(url, json=data, **kwargs)
        response.raise_for_status()
        logging.debug(f"RESPONSE: {response.status_code}")
        return response.json()

    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2) + wait_random(0, 2))
    def patch(self, url, data=None) -> Dict:
        """
        Perform a PATCH request against the REST server.
        """
        logging.debug(f"PATCH REQUEST: {url}")
        response = self._session.patch(url, json=data)
        response.raise_for_status()
        logging.debug(f"RESPONSE: {response.status_code}")
        return response

    @retry(stop=stop_after_attempt(3), wait=wait_fixed(2) + wait_random(0, 2))
    def delete(self, url, **kwargs) -> Dict:
        """
        Perform a DELETE request against the REST server.
        """
        logging.debug(f"DELETE REQUEST: {url}")
        response = self._session.delete(url, **kwargs)
        response.raise_for_status()
        logging.debug(f"RESPONSE: {response.status_code}")
        return response


class ModelResource:
    """
    Base class whose subclasses can used as the singlegon to bridge the local
    data model and the remote resource.
    """

    # the resource name/uri in the server
    resource_name = ""

    # the fields which are mandatory to create a resource in the server
    post_fields = []

    # the fields which are returned from a GET or POST request
    get_fields = []

    # the class which are used to create the local data model with remote
    # resource data
    model_class = None

    def __init__(self, server: RESTServer):
        self._server = server
        self._resource_url = join_url(self._server.url_root, self.resource_name)

    @property
    def resource_url(self):
        return self._resource_url

    def _validate_post_data(self, data: Dict) -> None:
        """
        Ensure all mandatory fields are provided to create the remote resource
        """
        for field in self.post_fields:
            if field not in data:
                raise FieldMissingError(field)

    def create_local_object(self, data: Dict) -> DataModel:
        """
        Create the local data model object with remote resource data
        """
        # pylint: disable=not-callable
        obj = self.model_class(model_resource=self)
        for field in self.get_fields:
            setattr(obj, field, data.get(field))
        obj.url = convert_http_to_https(obj.url)
        if hasattr(obj, "obj_id") and obj.obj_id:
            return obj
        obj.obj_id = get_id_from_url(obj.url)
        return obj

    def create_remote_resource(self, data) -> DataModel:
        """
        Create a remote resource.
        Return a local data model obj corresponding to the resource created.
        """
        self._validate_post_data(data)
        resource_data = self._server.post(self._resource_url, data)
        local_obj = self.create_local_object(resource_data)
        logging.info(f"Created: {local_obj.url}")
        return local_obj

    def delete_remote_resource(self, local_obj: DataModel) -> None:
        """
        Delete the remote resource.
        """
        self._server.delete(local_obj.url)
        logging.info(f"Deleted: {local_obj.url}")

    def get_remote_resource(self, local_obj: DataModel) -> DataModel:
        """
        Get the remote resource and return a new local data model object
        """
        resource_data = self._server.get(local_obj.url)
        return self.create_local_object(resource_data)

    def get_remote_resource_list(self, params=None) -> Generator:
        """
        Get the remote resource lists and return a generator for new local data model
        objects
        """
        resources = self._server.get_by_page(self._resource_url, params=params)
        return (self.create_local_object(resource) for resource in resources)


class ClientError(Exception):
    pass


class ServerError(Exception):
    pass


class FieldMissingError(ClientError):
    def __init__(self, field):
        error_msg = f"The field '{field}' is mandatory to create the resource."
        super().__init__(error_msg)


class MethodNotAllowedError(ClientError):
    pass
