import logging
import os
from enum import Enum

from cerberus import Validator

from sports.base import (
    DataModel,
    MethodNotAllowedError,
    ModelResource,
    RESTServer,
    ServerError,
)
from sports.ort_schemas import SCANNER_RESULT_SCHEMA
from sports.utils import join_url


class ScanFailedError(ServerError):
    def __init__(self, server_error_msg):
        error_msg = f"The scan has failed: {server_error_msg}"
        logging.error(error_msg)
        super().__init__(error_msg)


class ScanResultSchemaError(ServerError):
    def __init__(self, schema_error_msg):
        error_msg = f"The scan result's schema is invalid: {schema_error_msg}"
        logging.error(error_msg)
        super().__init__(error_msg)


class ScanStatus(str, Enum):
    NOT_STARTED = "not started"
    IN_PROGRESS = "in progress"
    COMPLETED = "completed"
    ERROR = "error"


class ORTServer(RESTServer):
    pass


class Project(DataModel):
    ...


class Version(DataModel):
    def create_code_location(self, file_path: str) -> str:
        """
        Shortcut to create a code location in the version. Actually, the work
        is done by the underlying model_resource
        """
        return self.model_resource.create_code_location(self, file_path)


class CodeLocation(DataModel):
    ...


class ScanResult(DataModel):
    ...


class ProjectResource(ModelResource):
    resource_name = "project"
    post_fields = ["name"]
    get_fields = ["name", "url"]
    model_class = Project


class VersionResource(ModelResource):

    resource_name = "version"
    post_fields = ["name"]
    get_fields = ["name", "url", "project"]
    model_class = Version

    def get_upload_url(self, version: Version) -> str:
        return join_url(version.url, "upload")

    def create_code_location(self, version: Version, file_path: str) -> str:
        if not os.path.isfile(file_path):
            raise Exception(f"The file path '{file_path}' does not exist.")
        file_name = os.path.basename(file_path)
        with open(file_path, "rb") as file_stream:
            files = {file_name: file_stream}
            url = self.get_upload_url(version)
            resp = self._server.post(url=url, files=files)
            if "success" not in resp or resp["success"] is False:
                raise Exception(f"Fail to upload the source: {resp}")
            code_loc_url = resp["url"]
            logging.info(f"Uploaded: {code_loc_url}")
            return code_loc_url


class CodeLocationResource(ModelResource):

    resource_name = "code-location"
    get_fields = ["filename", "version", "url"]
    model_class = CodeLocation

    def create_remote_resource(self, data):
        raise MethodNotAllowedError()


class ScanResultResource(ModelResource):

    resource_name = "scan-result"
    get_fields = ["url", "code_location", "status", "result", "error_msg"]
    model_class = ScanResult

    def create_remote_resource(self, data):
        raise MethodNotAllowedError()

    def delete_remote_resource(self, local_obj):
        raise MethodNotAllowedError()


class SchemaType(str, Enum):
    SCANNER = "scanner"


SCHEMAS = {SchemaType.SCANNER: SCANNER_RESULT_SCHEMA}


def get_schema_validator(
    schema_type: str, require_all=True, allow_unknown=True, **kwargs
) -> Validator:
    # We need to get the error msg if the validation fails. To avoid the overriding
    # between processes, we create a new validator each time.
    # All fields in schema are required. Extra unknown fields are allowed.
    schema = SCHEMAS.get(schema_type)
    return Validator(
        schema, require_all=require_all, allow_unknown=allow_unknown, **kwargs
    )
